/*
 * Copyright (C) 2016, 2017 Adrian Siekierka
 *
 * This file is part of FoamFix.
 *
 * FoamFix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FoamFix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FoamFix.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or
 * combining it with the Minecraft game engine, the Mojang Launchwrapper,
 * the Mojang AuthLib and the Minecraft Realms library (and/or modified
 * versions of said software), containing parts covered by the terms of
 * their respective licenses, the licensors of this Program grant you
 * additional permission to convey the resulting work.
 */

/**
 * Copyright (c) 2015 Kobata
 * Copyright (c) 2017 asiekierka
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package pl.asie.foamfix.client;

import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.texture.TextureUtil;
import org.lwjgl.opengl.ARBCopyImage;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL12;
import org.lwjgl.opengl.GL14;
import org.lwjgl.util.glu.GLU;
import pl.asie.foamfix.FoamFix;
import pl.asie.foamfix.api.IFoamFixSprite;
import pl.asie.foamfix.shared.FoamFixShared;
import pl.asie.foamfix.util.FoamUtilsClient;

import java.util.List;

public class FastTextureAtlasSprite extends TextureAtlasSprite implements IFoamFixSprite {
    private int textureId = -1;
    private int mipLevels = 0;

    public FastTextureAtlasSprite(String p_i1282_1_) {
        super(p_i1282_1_);
    }

    @Override
    public void func_94219_l() {
        if (FoamFixShared.config.clDisableTextureAnimations)
            return;

        ++field_110983_h;

        if (field_110983_h >= field_110982_k.func_110472_a(field_110973_g)) {
            int i = field_110982_k.func_110468_c(field_110973_g);
            int j = field_110982_k.func_110473_c() == 0 ? field_110976_a.size() : field_110982_k.func_110473_c();
            field_110973_g = (field_110973_g + 1) % j;
            field_110983_h = 0;
            int k = field_110982_k.func_110468_c(field_110973_g);

            if (i != k && k >= 0 && k < field_110976_a.size()) {

                if (textureId != -1) {
                    int destTex = GL11.glGetInteger(GL11.GL_TEXTURE_BINDING_2D);
                    FoamUtilsClient.checkGLError("updateAnimation | fastPath getPreviousTexture");

                    // Unbinding texture for safety, since copy image has an explicit destination.
                    GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);
                    FoamUtilsClient.checkGLError("updateAnimation | fastPath unbindTex");

                    for (int mip = 0; mip <= mipLevels; ++mip) {
                        ARBCopyImage
                                .glCopyImageSubData(textureId, GL11.GL_TEXTURE_2D, mip, (field_130223_c * k) >> mip, 0, 0, destTex, GL11.GL_TEXTURE_2D, mip, field_110975_c >> mip,
                                        field_110974_d >> mip,
                                        0, field_130223_c >> mip, field_130224_d >> mip, 1);
                        FoamUtilsClient.checkGLError("updateAnimation | fastPath mip="+mip);
                    }

                    GL11.glBindTexture(GL11.GL_TEXTURE_2D, destTex);
                    FoamUtilsClient.checkGLError("updateAnimation | fastPath rebindTex");
                } else {
                    int[][] frameData = this.field_110976_a.get(k);
                    uploadTextureMaxMips(mipLevels, frameData, field_130223_c, field_130224_d, field_110975_c, field_110974_d, false, false, frameData.length > 1);
                    FoamUtilsClient.checkGLError("updateAnimation | slowPath");
                }
            }
        } else if (this.field_110982_k.func_177219_e()) {
            this.func_180599_n();
        }
    }

    private int interpolateColor(double ratio, int from, int to) {
        return (int)((ratio * from) + ((1.0D - ratio) * to));
    }

    private boolean interpolateFrame(int[] to, int[] from1, int[] from2, double ratio) {
        if (from1.length == from2.length) {
            for (int i = 0; i < from1.length; ++i) {
                int color1 = from1[i];
                int color2 = from2[i];
                int colorRed = this.interpolateColor(ratio, color1 >> 16 & 0xFF, color2 >> 16 & 0xFF);
                int colorGreen = this.interpolateColor(ratio, color1 >> 8 & 0xFF, color2 >> 8 & 0xFF);
                int colorBlue = this.interpolateColor(ratio, color1 & 0xFF, color2 & 0xFF);
                to[i] = color1 & 0xFF000000 | colorRed << 16 | colorGreen << 8 | colorBlue;
            }
            return true;
        } else {
            return false;
        }
    }

    @Override
    protected void func_180599_n() {
        double ratio = 1.0D - (double)this.field_110983_h / (double)this.field_110982_k.func_110472_a(this.field_110973_g);
        int currentFrameIndex = this.field_110982_k.func_110468_c(this.field_110973_g);
        int frameCount = this.field_110982_k.func_110473_c() == 0 ? this.field_110976_a.size() : this.field_110982_k.func_110473_c();
        int nextFrameIndex = this.field_110982_k.func_110468_c((this.field_110973_g + 1) % frameCount);

        if (currentFrameIndex != nextFrameIndex && nextFrameIndex >= 0 && nextFrameIndex < this.field_110976_a.size()) {
            int[][] frame1 = this.field_110976_a.get(currentFrameIndex);
            int[][] frame2 = this.field_110976_a.get(nextFrameIndex);
            int mipLvl;
            if (this.field_176605_b == null || this.field_176605_b.length != mipLevels) {
                this.field_176605_b = new int[mipLevels][];
            }

            for (mipLvl = 0; mipLvl < mipLevels; ++mipLvl) {
                if (mipLvl >= frame1.length || mipLvl >= frame2.length) {
                    break;
                }

                if (this.field_176605_b[mipLvl] == null) {
                    this.field_176605_b[mipLvl] = new int[frame1[mipLvl].length];
                }

                interpolateFrame(this.field_176605_b[mipLvl], frame1[mipLvl], frame2[mipLvl], ratio);
            }

            uploadTextureMaxMips(mipLvl, this.field_176605_b, field_130223_c, field_130224_d, field_110975_c, field_110974_d, false, false, field_176605_b.length > 1);
        }
    }

    @Override
    public void func_110968_a(List<int[][]> textureData) {
        if(textureId != -1) {
            GL11.glDeleteTextures(textureId);
            textureId = -1;
            FoamUtilsClient.checkGLError("setFramesTextureData | deleteTexture");
        }

        super.func_110968_a(textureData);

        // TODO: Precalculate certain interpolated textures?

        // No need for extra texture if there's only one frame.
        // We're also not caching really long sets of animation.
        if (textureData.size() > 1 && textureData.size() <= FoamFixShared.config.txCacheAnimationMaxFrames && FoamFix.shouldFasterAnimation) {
            textureId = GL11.glGenTextures();
            FoamUtilsClient.checkGLError("setFramesTextureData | createTexture");

            int prevTex = GL11.glGetInteger(GL11.GL_TEXTURE_BINDING_2D);
            FoamUtilsClient.checkGLError("setFramesTextureData | getPreviousTexture");

            // Set up holding texture
            GL11.glBindTexture(GL11.GL_TEXTURE_2D, textureId);
            FoamUtilsClient.checkGLError("setFramesTextureData | bindTexture");

            GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL12.GL_TEXTURE_MAX_LEVEL, mipLevels);
            FoamUtilsClient.checkGLError("setFramesTextureData | setTextureProperties[Mip Levels]");
            GL11.glTexParameterf(GL11.GL_TEXTURE_2D, GL12.GL_TEXTURE_MIN_LOD, 0.0F);
            FoamUtilsClient.checkGLError("setFramesTextureData | setTextureProperties[LOD Min]");
            GL11.glTexParameterf(GL11.GL_TEXTURE_2D, GL12.GL_TEXTURE_MAX_LOD, (float) mipLevels);
            FoamUtilsClient.checkGLError("setFramesTextureData | setTextureProperties[LOD Max]");
            GL11.glTexParameterf(GL11.GL_TEXTURE_2D, GL14.GL_TEXTURE_LOD_BIAS, 0.0F);
            FoamUtilsClient.checkGLError("setFramesTextureData | setTextureProperties[LOD Bias]");

            // Reserve memory for texture
            for (int i = 0; i <= mipLevels; ++i) {
                GL11.glTexImage2D(GL11.GL_TEXTURE_2D, i, GL11.GL_RGBA, (field_130223_c * field_110976_a.size()) >> i, field_130224_d >> i, 0, GL12.GL_BGRA,
                        GL11.GL_UNSIGNED_BYTE, (java.nio.ByteBuffer) null);
                FoamUtilsClient.checkGLError("setFramesTextureData | createMip " + i);
            }

            // Copy animation frames to holding texture
            for (int i = 0; i < field_110976_a.size(); ++i) {
                uploadTextureMaxMips(mipLevels, field_110976_a.get(i), field_130223_c, field_130224_d, field_130223_c * i, 0, false, false, mipLevels > 0);
                FoamUtilsClient.checkGLError("setFramesTextureData | uploadFrame " + i);
            }

            // Restore old texture
            GL11.glBindTexture(GL11.GL_TEXTURE_2D, prevTex);
            FoamUtilsClient.checkGLError("setFramesTextureData | restoreTexture");
        }
    }

    private static void uploadTextureMaxMips(int maxMips, int[][] data, int width, int height, int originX, int originY, boolean linearFiltering, boolean clamped, boolean mipFiltering) {
        int mips = maxMips >= 0 ? Math.min(maxMips, data.length-1) : data.length-1;
        for (int mip = 0; mip <= mips; ++mip) {
            if ((width >> mip) > 0 & (height >> mip) > 0) {
                TextureUtil.func_147947_a(mip, data[mip], width >> mip, height >> mip, originX >> mip, originY >> mip, linearFiltering,
                        clamped, mipFiltering);
                FoamUtilsClient.checkGLError("uploadTextureMaxMips mip=" + mip);
            }
        }
    }

    @Override
    public void func_130103_l() {
        super.func_130103_l();

        if(textureId != -1) {
            GL11.glDeleteTextures(textureId);
            textureId = -1;
        }
    }

    @Override
    public void func_147963_d(int p_147963_1_) {
        mipLevels = FoamFixShared.config.txMaxAnimationMipLevel >= 0 ? Math.min(FoamFixShared.config.txMaxAnimationMipLevel, p_147963_1_) : p_147963_1_;

        super.func_147963_d(p_147963_1_);
    }

    @Override
    public boolean isStoredOnGPU() {
        return textureId != -1 || !func_130098_m();
    }
}
