/*
 * Copyright (C) 2016, 2017 Adrian Siekierka
 *
 * This file is part of FoamFix.
 *
 * FoamFix is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FoamFix is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FoamFix.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify this Program, or any covered work, by linking or
 * combining it with the Minecraft game engine, the Mojang Launchwrapper,
 * the Mojang AuthLib and the Minecraft Realms library (and/or modified
 * versions of said software), containing parts covered by the terms of
 * their respective licenses, the licensors of this Program grant you
 * additional permission to convey the resulting work.
 */

package pl.asie.foamfix.common;

import com.google.common.collect.*;
import net.minecraft.block.Block;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;

import java.util.Map;

public class FoamyBlockState extends BlockStateContainer.StateImplementation {
	protected final PropertyValueMapper owner;
	protected final ImmutableMap<IProperty<?>, Comparable<?>> properties;
	protected int value;

	public FoamyBlockState(PropertyValueMapper owner, Block blockIn, ImmutableMap < IProperty<?>, Comparable<? >> propertiesIn) {
		super(blockIn, propertiesIn);
		this.owner = owner;
		this.properties = propertiesIn;
	}

	@Override
	public <T extends Comparable<T>, V extends T> IBlockState func_177226_a(IProperty<T> property, V value) {
		Comparable<?> comparable = this.properties.get(property);

		if (comparable == null) {
			throw new IllegalArgumentException("Cannot set property " + property + " as it does not exist in " + this.func_177230_c().func_176194_O());
		} else if (comparable == value) {
			return this;
		} else {
			IBlockState state = owner.withProperty(this.value, property, value);

			if (state == null) {
				throw new IllegalArgumentException("Cannot set property " + property + " to " + value + " on block " + Block.field_149771_c.func_177774_c(this.func_177230_c()) + ", it is not an allowed value");
			} else {
				return state;
			}
		}
	}

	@Override
	public void func_177235_a(Map <Map< IProperty<?>, Comparable<? >>, BlockStateContainer.StateImplementation > map) {
		this.value = owner.generateValue(this);
	}
}
